#include "settings.hh"
#include "x.hh"

#include <unistd.h> /* usleep */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xmu/Xmu.h>

unsigned width, height, depth;

static Display *mDisplay;
static int mScreen;
static Window mRootWin, window;
static XVisualInfo vinfo;

static XImage *image;

#ifdef USE_SHM
 #include <sys/ipc.h>
 #include <sys/shm.h>   
 #include <X11/extensions/XShm.h>

 static XShmSegmentInfo Shminfo;
 static bool Using_SHM;
#endif

#ifdef USE_XV
 #include <X11/extensions/Xvlib.h>
 
 static XvPortID XVport;
 static bool Using_XV;
#endif

static int Error(Display *dpy, XErrorEvent *err)
{
  (void) XmuPrintDefaultErrorMessage (dpy, err, stderr);
  return 0;
}

static void Failed(const char *what)
{
    fprintf(stderr, "%s failed...\n", what);
    fflush(stderr);
}

static Window XWindowByName(Display *display, const Window rootwin, const char *name)
{
    unsigned int num_children;
    Window *children, child, window;
    XTextProperty windowname;

    if(XGetWMName(display, rootwin, &windowname) != 0)
    {
        fprintf(stderr, "Window='%s'\n", (const char *)windowname.value);
        if(!strcmp((const char *)windowname.value, name))
            return rootwin;
    }
    
    window = (Window) NULL;
     
    if(XQueryTree(display, rootwin, &child, &child, &children, &num_children))
    {
        unsigned i;
        for(i=0; i < num_children; ++i)
        {
            /* Search each child and their children. */
            window = XWindowByName(display, children[i], name);
            if(window != (Window) NULL)
                break;
        }
        if (children != (Window *)NULL)
            XFree((void *)children);
    }
    return window;
}
static void XListWindows(Display *display, const Window rootwin)
{
    unsigned int num_children;
    Window *children, child;
    XTextProperty windowname;

    if(XGetWMName(display, rootwin, &windowname) != 0)
        printf("  '%s'\n", (const char *)windowname.value);
    
    if(XQueryTree(display, rootwin, &child, &child, &children, &num_children))
    {
        unsigned i;
        for(i=0; i < num_children; ++i)
        {
            /* Search each child and their children. */
            XListWindows(display, children[i]);
        }
        if (children != (Window *)NULL)
            XFree((void *)children);
    }
}

static void GetWindowGeometry()
{
    unsigned border_width;
    int xpos, ypos;
    Window root;
    
    fprintf(stderr, "XGetGeometry...\n");
    if(!XGetGeometry(
                 mDisplay, window, &root,
                 &xpos, &ypos,
                 &width, &height,
                 &border_width,
                 &depth))
    {
        Failed("XGetGeometry");
    }
    fprintf(stderr, "width=%u, height=%u, depth=%u\n", width, height, depth);
    
    width  -= XOFFSET;
    height -= YOFFSET;
}

void XInit()
{
    char *dispName = XDisplayName(NULL);

    XSetErrorHandler(Error);

    mDisplay = XOpenDisplay(dispName);
    if(!mDisplay)
    {
        fprintf(stderr, "XOpenDisplay(%s) failed\n", dispName);
        return;
    }
    mScreen = DefaultScreen(mDisplay);
    
    mRootWin = RootWindow(mDisplay, mScreen);
    
    for(;;)
    {
        window = XWindowByName(mDisplay, mRootWin, WINDOW_NAME.c_str());
        if(window) break;
        fprintf(stderr, "--Retrying--\n");
        usleep(20000);
    }
    
    XFlush(mDisplay);
    XSync(mDisplay, False);
    
    GetWindowGeometry();

#ifdef USE_SHM
    {
        Using_SHM = XShmQueryExtension(mDisplay);
        if(!Using_SHM)
            Failed("XShmQueryExtension");
    }
#endif

#ifdef USE_XV
    {
        unsigned ver,rel,req,ev,err;
        Using_XV = XvQueryExtension(mDisplay, &ver,&rel,&req,&ev,&err) == Success;
        if(!Using_XV)
        {
            Failed("XvQueryExtension");
        }
        
        if(Using_XV)
        {
            unsigned adaptors;
            XvAdaptorInfo *ai;
            XvQueryAdaptors(mDisplay, mRootWin, &adaptors, &ai);
            
            Using_XV = false;
            for(unsigned a=0; a<adaptors; ++a)
                if(ai[a].type & (XvOutputMask | XvVideoMask))
                {
                    fprintf(stderr, "Considering XV adapter %u\n", a);
                    for(unsigned b=0; b<ai[a].num_ports; ++b)
                    {
                        XVport = ai[a].base_id + b;
                        if(!XvGrabPort(mDisplay, XVport, CurrentTime))
                        {
                            fprintf(stderr, "Using XV port %d\n", XVport);
                            Using_XV = true;
                        }
                    }
                }
        }
            
        if(Using_XV)
        {
            Using_XV = false;
            
            unsigned formats;
            XvImageFormatValues *fo = XvListImageFormats(mDisplay, XVport, (int *)&formats);
            for(unsigned a=0; a<formats; ++a)
                if(fo[a].type == XvRGB)
                    Using_XV = true;
            
            if(!Using_XV) fprintf(stderr, "No suitable colour format, not using XV\n");
        }
        
        if(Using_XV)
        {
            XGCValues xgcv;
            GC gc = XCreateGC(mDisplay, window, 0, &xgcv);
        
            XvGetVideo(mDisplay, XVport, window, gc,
                       0,0,             width,height,
                       XOFFSET,YOFFSET, width,height);
        }
    }
#endif

    if(!XMatchVisualInfo(mDisplay, mScreen, depth, DirectColor, &vinfo))
    {
        Failed("XMatchVisualInfo");
        return;
    }
}

unsigned char *GetImage()
{
#ifdef USE_XV
#endif

#ifdef USE_SHM
    if(!image && Using_SHM)
    {
        fprintf(stderr, "XShmCreateImage...\n");
        image = XShmCreateImage(mDisplay, vinfo.visual,
                  depth,
                  ZPixmap,
                  NULL,
                  &Shminfo,
                  width,
                  height);
        if(!image)
        {
            Failed("XShmCreateImage");
        CancelSHM:
            Using_SHM = false;
            return GetImage();
        }
        
        fprintf(stderr, "shmget...\n");
        Shminfo.shmid = shmget(IPC_PRIVATE,
                               image->bytes_per_line * image->height,
                               IPC_CREAT | 0777);
        if(Shminfo.shmid < 0)
        {
            perror("shmget");
        CancelSHM2:
            XDestroyImage(image); image=NULL;
            goto CancelSHM;
        }
        fprintf(stderr, "shmat...\n");
        Shminfo.shmaddr = (char *) shmat(Shminfo.shmid, 0, 0);
        if(Shminfo.shmaddr == (char *)-1)
        {
            perror("shmat");
            goto CancelSHM2;
        }
        fprintf(stderr, "shmctl...\n");
        shmctl(Shminfo.shmid, IPC_RMID, 0),

        image->data = Shminfo.shmaddr;
        Shminfo.readOnly = False;
        fprintf(stderr, "XShmAttach...\n");
        if(!XShmAttach(mDisplay, &Shminfo))
        {
            Failed("XShmAttach");
            goto CancelSHM2;
        }
        
        XSync(mDisplay, False);
    }
#endif

    /* If SHM failed or was disabled, try non-SHM way */
    if(!image)
    {
        /*fprintf(stderr, "XGetImage...\n");
        fflush(stderr);*/
        
        image = XGetImage(mDisplay, window,
                  XOFFSET, YOFFSET,
                  width,
                  height,
                  AllPlanes,
                  XYPixmap);
        if(!image)
        {
            Failed("XGetImage");
        }
    }
    
#ifdef USE_SHM
    if(Using_SHM)
    {
        /*fprintf(stderr, "XSHMGetImage...\n");
        fflush(stderr);*/
        if(!XShmGetImage(mDisplay, window, image,
                        XOFFSET, YOFFSET,
                        AllPlanes))
        {
            Failed("XSHMGetImage");
        }
    }
#endif
    return image ? (unsigned char *)image->data : NULL;
}

void DeleteImage(void)
{
    if(!image) return;

#ifdef USE_SHM
    if(Using_SHM) return;
#endif
    fprintf(stderr, "Destroying image\n");
    fflush(stderr);
    
    fprintf(stderr, "XDestroyImage...\n");
    XDestroyImage(image);
    image = NULL;
}

void XDone(void)
{
    if(!image) return;
    
#ifdef USE_SHM
    if(Using_SHM)
    {
        fprintf(stderr, "Destroying image\n");
        fflush(stderr);

        fprintf(stderr, "XShmDetach...\n");
        XShmDetach(mDisplay, &Shminfo);
        fprintf(stderr, "XDestroyImage...\n");
        XDestroyImage(image);
        image = NULL;
        fprintf(stderr, "shmdt...\n");
        shmdt(Shminfo.shmaddr);
    }
#endif
}
